#!/usr/bin/perl -w
# $Id: ace2xml.pl 100 2007-08-06 23:41:26Z pjohnson $

# ace2xml.pl -- Takes ACE format (from phred/phrap/consed) + quality
# scores in FASTA format and converts to XML conforming to piim.dtd
#
# Created 2 May 2007 by Philip Johnson

use strict;
use Getopt::Std;

{
    my %opts;
    $opts{o} = '';
    if (!getopts("i:q:o:", \%opts)  ||  !defined $opts{i}  ||
        !defined $opts{q}) {
        print "usage: $0 -i <input.ace> -q <input.qual> [-o <output-file-prefix>]\n",
            "\t<input.qual> should contain fasta-esque formatting of
        quality scores for each read used in the ACE file.  XML for
        each contig will be placed in a separate file named
        '<outfile-file-prefix><contig-name>.xml'\n\n";
        exit 1;
    }


    open QUALFILE, $opts{q};
    my $qualIndex = CreateQualFileIndex(\*QUALFILE);

    open ACEFILE, $opts{i} or  die "Error opening ACE file: '$opts{i}' ($!)\n";
    my ($numContigs) = <ACEFILE> =~ /^AS\s+(\d+)\s/;
    #die "$0 can only handle 1 contig per ace file\n" if ($numContigs > 1);
    
    for (my $i = 0;  $i < $numContigs;  ++$i) {
        while (defined($_ = <ACEFILE>)  &&  !/^CO/) {}
        ProcessContig(\*ACEFILE, \*QUALFILE, $qualIndex, $opts{o});
    }

    close ACEFILE;
    close QUALFILE;
}

# -----------------------------------------------------------------------------
# Takes ACEFILE queued to start of contig, along with QUALFILE & index
#
sub ProcessContig {
    my ($ACEFILE, $QUALFILE, $qualIndex, $outfilePre) = @_;

    my ($ctgName, $numBases, $numReads) = /^CO\s+(\S+)\s+(\d+)\s+(\d+)/;

    while (defined($_ = <$ACEFILE>)  &&  !/^AF/) {} #skip consensus

    my @assm;
    do {
        my ($acc, $ori, $pos) = /^AF\s+(\S+)\s+(U|C)\s+(-?\d+)/;
        #switch to 0-based coordinates to maintain sanity
        push @assm, [$acc, $ori eq 'C', $pos-1];
    } while (defined($_ = <$ACEFILE>)  &&  /^AF/);


    #main loop over ctg
    my @currReads;
    my $assmPos = 0;
    open OUTFILE, '>'.$outfilePre.$ctgName.'.xml';
    print OUTFILE '<?xml version="1.0"?>', "\n";
    print OUTFILE '<!DOCTYPE Contig PUBLIC "" "piim.dtd">', "\n";
    print OUTFILE "<Contig>\n";
    for (my $ctgPos = 0;  $ctgPos < $numBases;  ++$ctgPos) {
        print OUTFILE "  <Column pos=\"$ctgPos\">\n";
        LoadReadsFor(\@assm, \$assmPos, $ctgPos, $ACEFILE,
                     $QUALFILE, $qualIndex, \@currReads);
        my $labels = AssignLabels2Nts(\@currReads, $ctgPos);
        foreach my $read (@currReads) {
            my $p = $ctgPos - $read->[0];
            if ($read->[2][$p] !~ /A|C|G|T/) { #screen uncalled bases
                next;
            }

            print OUTFILE '    <Chr score="', $read->[3][$p], '" ', $labels->{$read->[2][$p]}, "/>\n";
        }
        print OUTFILE "  </Column>\n";
    }
    print OUTFILE "</Contig>\n";
    close OUTFILE;
}


# -----------------------------------------------------------------------------
# Create hash from accession --> position in qual file
#
sub CreateQualFileIndex {
    my ($QUALFILE) = @_;

    my $qualIndex = {};
    while (<$QUALFILE>) {
        if (/^>/) {
            my ($acc) = /^>(.+?) /;
            $qualIndex->{$acc} = tell $QUALFILE;
        }
    }

    return $qualIndex;
}

# -----------------------------------------------------------------------------
# Read and concatenate lines (presumably nucleotides) until hitting a
# blank line
#
sub ReadBases {
    my ($ACEFILE) = @_;

    my $seq = [];
    while (defined($_ = <$ACEFILE>)  &&  !/^$/) {
        chomp;
        push @$seq, split(//, uc $_);
    }

    return $seq;
}

# -----------------------------------------------------------------------------
# Read and concatenate q-scores for a single read
#
sub ReadQuals {
    my ($QUALFILE) = @_;

    my $quals = [];
    while (defined($_ = <$QUALFILE>)  &&  !/^>/) {
        chomp;
        push @$quals, split(/ /, $_);
    }

    return $quals;
}

# -----------------------------------------------------------------------------
# Assign labels "minor", "other1" and "other2" to different nucleotides
#
sub AssignLabels2Nts {
    my ($reads, $ctgPos) = @_;
    my $labels = {'A'=>0, 'C'=>0, 'G'=>0, 'T'=>0};

    foreach my $read (@$reads) {
        my $p = $ctgPos - $read->[0];
        if ($read->[2][$p] !~ /A|C|G|T/) { #screen uncalled bases
            next;
        }
        ++($labels->{$read->[2][$p]});
    }

    my @sorted = (sort({$labels->{$b} <=> $labels->{$a}} keys(%$labels)));
    $labels->{$sorted[0]} = '';
    $labels->{$sorted[1]} = 'minor=""';
    $labels->{$sorted[2]} = 'other="1"';
    $labels->{$sorted[3]} = 'other="2"';
    return $labels;
}

# -----------------------------------------------------------------------------
# Load reads (bases & q-scores) that cover the current position (ctgPos)
#
sub LoadReadsFor {
    my ($assm, $assmPos, $ctgPos, $ACEFILE, $QUALFILE, $qualIndex, $reads)
        = @_;

    #delete stale reads
    for (my $i = 0;  $i < @$reads;  ) {
        if ($reads->[$i][0]+$reads->[$i][1] <= $ctgPos) {
            splice @$reads, $i,1; #delete!
        } else {
            ++$i;
        }
    }

    #load new ones
    for (;  $$assmPos < @$assm  &&  $assm->[$$assmPos][2] <= $ctgPos;
         ++$$assmPos) {
        while (defined($_ = <ACEFILE>)  &&  !/^RD/) {}
        my ($acc, $paddedLength) = /^RD\s+(\S+)\s+(\d+)/;
        die "Error: order mismatch between assembly (AF:$assm->[$$assmPos][0]) and reads (RD:$acc)\n"
            if $acc ne $assm->[$$assmPos][0];

        my $bases = ReadBases($ACEFILE); #already complemented if necessary

        my $qualPos = $qualIndex->{$acc};
        die "Error: cannot find qualities for $acc\n" if (!defined($qualPos));
        seek $QUALFILE, $qualIndex->{$acc}, 0;
        my $quals = ReadQuals($QUALFILE);
        if ($assm->[$$assmPos][1]) { #complement -- so flip quals
            @$quals = reverse(@$quals);
        }
        for (my $i = 0;  $i < @$bases;  ++$i) { #insert gaps into quals
            if ($bases->[$i] eq '*') {
                splice @$quals, $i, 0, 0;
            }
        }

        die "Error: base call & q-score mismatch for $acc" if scalar(@$bases) != scalar(@$quals);
        push @$reads, [$assm->[$$assmPos][2], $paddedLength, $bases, $quals];
    }
}
