/* $Id: gsl_util.hpp 61 2006-07-31 21:15:25Z pjohnson $
  ------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2006. "Inference of population
    genetic parameters in metagenomics: a clean look at messy data."
    Genome Research (in press).
 
  Copyright 2006 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Generic wrappers for GSL functions that take a callbacks
  ---------------------------------------------------------------------------
*/
#ifndef GSL_UTIL_HPP
#define GSL_UTIL_HPP

#include <iostream>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_siman.h>
#include <gsl/gsl_roots.h>

using namespace std;

//---------------------------------------------------------------------------

class CGSL_NIntegrate {
public:
    typedef CGSL_NIntegrate TThisType;
    virtual ~CGSL_NIntegrate(void) {} //virtual to silence warning

    virtual double F(double x) const = 0;
    double Integrate(double from, double to) const {
        double result, error;
        gsl_function f;
        f.function = &s_F;
        f.params = (void*) this;

        gsl_integration_workspace *w = gsl_integration_workspace_alloc(10000);
        gsl_integration_qag(&f, from, to,
                            0, 1e-7, 10000, GSL_INTEG_GAUSS15,
                            w, &result, &error);
        gsl_integration_workspace_free(w);
        return result;
    }

protected:
    //static callback functions
    static double s_F(double x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }    
};

//---------------------------------------------------------------------------

class CGSLSimAnneal {
public:
    virtual ~CGSLSimAnneal(void) {}
    virtual double E(void) const = 0;
    virtual void TakeStep(const gsl_rng *r, double maxSize) = 0;
    virtual double DistanceFrom(const CGSLSimAnneal *sa) const = 0;
    virtual void PrintCfg(void) const = 0;
    virtual CGSLSimAnneal& operator= (const CGSLSimAnneal &sa) = 0;
    virtual CGSLSimAnneal* Clone(void) const = 0;

    void Solve(const gsl_siman_params_t &params) {
        gsl_rng * r;
        
        gsl_rng_env_setup();
        r = gsl_rng_alloc(gsl_rng_default);
        
        gsl_siman_solve(r, this, s_E, s_Step, s_Dist, s_Print,
                        s_Copy, s_Construct, s_Destruct, 0, params);
    }

protected:
    typedef CGSLSimAnneal TThisType;
    //static callback functions
    static double s_E(void *cls) {
        return ((const TThisType*) cls)->E();
    }
    static void s_Step(const gsl_rng *r, void *cls, double maxStep) {
        ((TThisType*) cls)->TakeStep(r, maxStep);
    }
    static double s_Dist(void *cls1, void *cls2) {
        return ((const TThisType*) cls1)->DistanceFrom((const TThisType*)cls2);
    }
    static void s_Print(void *cls) {
        ((const TThisType*) cls)->PrintCfg();
    }
    static void s_Copy(void *src, void *dst) {
        *((TThisType*) dst) = *((const TThisType*) src);
    }
    static void* s_Construct(void *cls) {
        return ((const TThisType*) cls)->Clone();
    }
    static void s_Destruct(void *cls) {
        delete ((TThisType*) cls);
    }
    //end static functions
};

//---------------------------------------------------------------------------

class CGSLMultiMin {
public:
    typedef CGSLMultiMin TThisType;
    typedef gsl_vector TParam;

    virtual ~CGSLMultiMin(void) {} //virtual to silence warning
    virtual double F(const TParam*) const = 0;
    virtual void dF(const TParam *, TParam *) const {};
    virtual void FdF(const TParam *x, double *f, TParam *g) const {
        *f = F(x); dF(x, g);
    }

protected:
    //static callback functions
    static double s_F(const TParam *x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }
    static void s_dF(const TParam *x, void *cls, TParam *g) {
        ((const TThisType*) cls)->dF(x, g);
    }
    static void s_FdF(const TParam *x, void *cls, double *f, TParam *g) {
        ((const TThisType*) cls)->FdF(x, f, g);
    }
    //end static functions
};

//---------------------------------------------------------------------------

class CGSLSimplex : CGSLMultiMin {
protected:
    void Optimize(double stepSize, double tolerance, unsigned int maxIter);
    bool HasConverged(void) const { return m_Converged; }

    virtual unsigned int x_GetNumParams(void) const = 0;
    virtual void x_InitializeParams(TParam*) const = 0;
    virtual void x_StoreParams(const TParam*) = 0;
private:
    bool m_Converged;
};

/*---------------------------------------------------------------------------*/
inline void CGSLSimplex::Optimize(double stepSize, double tolerance,
                                  unsigned int maxIter) {
    gsl_multimin_function func;

    func.f = &s_F;
    func.params = (void*) this;
    func.n = x_GetNumParams();

    gsl_vector *stepVector = gsl_vector_alloc(func.n);
    gsl_vector_set_all(stepVector, stepSize);

    gsl_vector *x = gsl_vector_alloc(func.n);
    x_InitializeParams(x);

    gsl_multimin_fminimizer *minimizer =
        gsl_multimin_fminimizer_alloc(gsl_multimin_fminimizer_nmsimplex,func.n);
    gsl_multimin_fminimizer_set(minimizer, &func, x, stepVector);

    int status;
    unsigned int i = 0;
    do {
        ++i;
        gsl_multimin_fminimizer_iterate(minimizer);
        status = gsl_multimin_test_size(gsl_multimin_fminimizer_size(minimizer),
                                        tolerance);
#ifdef _DEBUG
        cerr << "Optimization iteration #" << i << "\n";
        cerr << gsl_multimin_fminimizer_minimum(minimizer) << "\t";
        gsl_vector *p = gsl_multimin_fminimizer_x(minimizer);
        for (unsigned int i = 0;  i < p->size;  ++i) {
            cerr << gsl_vector_get(p, i) << " ";
        }
        cerr << endl;
#endif
    } while (status == GSL_CONTINUE  &&  i < maxIter);
    
    m_Converged = (status == GSL_SUCCESS);

    //assign back regardless of whether we've achieved GSL_SUCCESS
    x_StoreParams(gsl_multimin_fminimizer_x(minimizer));

    gsl_multimin_fminimizer_free(minimizer);
    gsl_vector_free(x);
    gsl_vector_free(stepVector);
}

//---------------------------------------------------------------------------

class CGSLFunction {
public:
    typedef CGSLFunction TThisType;

    virtual ~CGSLFunction(void) {} //virtual to silence warning
    virtual double F(const double) const = 0;

protected:
    //static callback functions
    static double s_F(const double x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }
    //end static functions

    const gsl_function x_GetGSLFunc(void) const {
        gsl_function f;
        f.function = &s_F;
        f.params = (void*) this;
        return f;
    }
};

class CGSLFindRootBracket : public CGSLFunction {
protected:
    double FindRoot(double lowerBound, double upperBound,
                    double tolerance, unsigned int maxIter)
    {
#ifdef _DEBUG
        cerr << "lb: " << lowerBound << "\tub: " << upperBound << endl;
#endif
        gsl_function func(x_GetGSLFunc());
        gsl_root_fsolver *solver =
            gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
        gsl_root_fsolver_set(solver, &func, lowerBound, upperBound);

        int status;
        unsigned int i = 0;
        do {
            gsl_root_fsolver_iterate(solver);
            status = gsl_root_test_interval(gsl_root_fsolver_x_lower(solver),
                                            gsl_root_fsolver_x_upper(solver),
                                            0, tolerance);
        } while(status == GSL_CONTINUE  &&  ++i < maxIter);

        m_Converged = (status == GSL_SUCCESS);

        double root = gsl_root_fsolver_root(solver);
        gsl_root_fsolver_free(solver);

        return root;
    }

    bool HasConverged(void) const { return m_Converged; }

private:
    bool m_Converged;
};

//---------------------------------------------------------------------------


#endif /*ifndef GSL_UTIL_HPP*/
