#!/usr/bin/perl -w

# ms2spectrum.pl -- this script simulates the frequency spectra from a
#metagenomic-style contig.  It requires Richard Hudson's program
#"m.s." (http://home.uchicago.edu/~rhudson1/source/mksamples.html).
#Two changes to the source of m.s. may be advisable, depending on your
#intended use: 1) increase the precision of the real number giving the
#position of each segregating site (as a rule of thumb, it should have
#at least as many digits as sites in your simulated contig) and 2) if
#you plan to do many simulations in parallel, use a random number seed
#algorithm that handles this case properly.

use strict;
use Getopt::Std;

#-------------------------------------------------------------------
# Assume reads are randomly distributed across contig
sub GenerateReadStarts($$$) {
    my ($depth, $readLength, $numSites) = @_;
    my @starts;
    
    my $numReads = int($depth * ($numSites / $readLength));
    for (my $i = 0;  $i < $numReads;  ++$i) {
        push @starts, int(rand($numSites - $readLength));
    }

    @starts = sort @starts;
    return @starts;
}

#-------------------------------------------------------------------
# Convert m.s.'s [0,1] real coords to our integer coords [0,$numSites)
# Also tweak so segregating sites don't land on top of each other
# (helps to adjust the m.s. source code to output numbers with more
# than the default 4 digits of precision.
sub AdjustSeqPos($$) {
    my ($seqPos, $numSites) = @_;

    for (my $i = 1;  $i < @$seqPos;  ++$i) {
        $seqPos->[$i] = int($seqPos->[$i]*$numSites);
        if ($seqPos->[$i] <= $seqPos->[$i-1]  &&
            ($seqPos->[$i-1]+1) < $numSites) {
            $seqPos->[$i] = $seqPos->[$i-1]+1;
        }
    }
}

#-------------------------------------------------------------------

sub RunMS($$$$$) {
    my ($aveDepth, $numSites, $theta, $growth, $rho) = @_;
    my $readLength = 800.0;
    my @spectrum;

    my @readStarts = GenerateReadStarts($aveDepth, $readLength, $numSites);
    my $numReads = @readStarts;

    $theta *= $numSites; #convert to per-locus theta
    open(MS, "ms $numReads 1 -t $theta -r $rho $numSites -G $growth |") or die;
    while (defined ($_ = <MS>)  &&  !/^segsites:/) {}
    my ($segSites) = /^segsites: (\d+)/;
    if ($segSites > 0) {
        $_ = <MS>;
        ($_) = /^positions: (\d.+)$/;
        my @seqPos = split;
        AdjustSeqPos(\@seqPos, $numSites);

        my @ctgDerived = (0) x $numSites;
        my @ctgDepth = (0) x $numSites;

        my $sampleNum = 0;
        while (<MS>) { #one line for each sample ('depth' in total)
            chomp;
            #string of 0 and 1 for ancestral / derived.  length = segsites
            my @haplotype = split //;

            #add this sample to depth counters
            for (my $pos = $readStarts[$sampleNum];
                 $pos < $readStarts[$sampleNum] + $readLength;
                 ++$pos) {
                ++$ctgDepth[$pos];
            }

            #add this sample to derived counters
            for (my $i = 0;  $i < @haplotype;  ++$i) {
                if ($seqPos[$i] >= $readStarts[$sampleNum]  &&
                    $seqPos[$i] < $readStarts[$sampleNum]+$readLength) {
                    if ($i > 0  &&  $seqPos[$i] == $seqPos[$i-1]) {
                        #print "WARNING: $seqPos[$i] and $seqPos[$i-1]\n";
                        next;
                    }
                    $ctgDerived[$seqPos[$i]] += $haplotype[$i];
                }
            }
            ++$sampleNum;
        }


        #populate spectrum
        for (my $i = 0;  $i < @ctgDepth;  ++$i) {
            if ($ctgDepth[$i] > 0  &&  $ctgDerived[$i] == $ctgDepth[$i]) {
                $ctgDerived[$i] = 0; #can't distinguish derived/ancestral
            }
            ++$spectrum[$ctgDepth[$i]][$ctgDerived[$i]];
        }
        
        #zero where necessary
        for (my $i = 0;  $i < @spectrum;  ++$i) {
            if (defined $spectrum[$i]) {
                for (my $j = 0;  $j < $i;  ++$j) {
                    $spectrum[$i][$j] = 0 if !defined ($spectrum[$i][$j]);
                }
            }
        }

        if (@spectrum > 20) { #cap depth at 20 for speed
            $#spectrum = 20;
        }

        #normalize
        my $tot = 0;
        for (my $i = 1;  $i < @spectrum;  ++$i) {
            if (defined $spectrum[$i]) {
                for (my $j = 0;  $j < $i;  ++$j) {
                    $tot += $spectrum[$i][$j];
                }
            }
        }
        for (my $i = 1;  $i < @spectrum;  ++$i) {
            if (defined $spectrum[$i]) {
                for (my $j = 0;  $j < $i;  ++$j) {
                    $spectrum[$i][$j] /= $tot;
                }
            }
        }
    }
    close(MS);

    return @spectrum;
}


#-----MAIN--------
my %opts;
if (!getopts("d:n:t:g:r:", \%opts)  ||  keys(%opts) != 5) {
    print "Syntax: $0 -d <aveDepth> -n <# sites> -t <theta> -g <growth> -r <recomb> \n\n";
    exit 1;
}

my @fs = RunMS($opts{d}, $opts{n}, $opts{t}, $opts{g}, $opts{r});

for (my $i = 1;  $i < @fs;  ++$i) {
    if (defined $fs[$i]) {
        print $i, "\n";
        print join(' ', @{$fs[$i]}), "\n";
    }
}
